#lang scheme

(require "structures.scm")

(provide parse-file parse-files parse-program
         parse-body
         parse-expr
         parse-decl
         )

; decls
(define (parse-file path)
  (parse-files (list path)))

(define (parse-files paths)
  (flatten (map (lambda(path)
                  (call-with-input-file path (lambda(in)
                                               (parse-program (read in)))))
                paths)))

(define (parse-program sexpr)
  (map parse-decl sexpr))

(define (parse-decl sexpr)
  (let ((members (parse-members (cdddr sexpr))))
    (make-class-decl (cadr sexpr)
                     (caddr sexpr)
                     (filter field-decl?  members)
                     (filter method-decl? members))))

(define (parse-members sexprs) (map parse-member sexprs))

(define (parse-member sexpr)
  (case (car sexpr)
    ((field)  (make-field-decl (cadr sexpr)))
    ((method) (make-method-decl (cadr sexpr) 
                                (caddr sexpr) 
                                (parse-body (cdddr sexpr))
                                #f))
    ((sync-method) (make-method-decl (cadr sexpr) 
                                     (caddr sexpr) 
                                     (parse-body (cdddr sexpr))
                                     #t))
    (else (error 'parse-member "not valid decl: ~a~n" sexpr))
    ))

; 
(define (parse-body sexprs)  (parse-exprs sexprs))
(define (parse-exprs sexprs) 
  ;(printf "exprs: ~a~n" sexprs) 
  (map (lambda(s)
         ;(printf "s: ~a~n" s)
         (parse-expr s)) sexprs))

; exprs
(define (parse-expr sexpr)
  ;(printf "parse: ~a~n" sexpr)
  (cond 
    ((prim-lit? sexpr) (make-lit-expr sexpr))
    ((list-lit? sexpr) (make-lit-expr (cdr sexpr)))
    ((symbol?   sexpr) (make-vget-expr sexpr))
    (else
     (case (car sexpr)
       ; quoting 
       ((quote) (make-lit-expr (cadr sexpr)))
       
       ((set!) (make-vset-expr (cadr sexpr) 
                               (parse-expr (caddr sexpr))))
       
       ((and) (make-and-expr (parse-exprs (cdr sexpr))))
       
       ((or) (make-or-expr (parse-exprs (cdr sexpr))))
       
       ((cond) (make-cond-expr 
                (map (lambda(c)
                       (list (if (equal? (car c) 'else)
                                 (parse-expr #t)
                                 (parse-expr (car c)))
                             (parse-expr (cadr c))))
                     (cdr sexpr))))
       
       ; if = syntactic sugar for a "cond"
       ((if) (parse-expr `(cond (,(cadr sexpr) ,(caddr sexpr))
                                (else ,(cadddr sexpr)))))
       
       ; (let (()) body 
       ((let)  (make-let-expr (map car (cadr sexpr))
                              (parse-exprs (map cadr (cadr sexpr)))
                              (parse-exprs (cddr sexpr))))
       
       ; while
       ((while) (make-while-expr (parse-expr (cadr sexpr))
                                 (map parse-expr (cddr sexpr))))
       ; new
       ((new)  (make-new-expr (cadr sexpr)
                              (parse-exprs (cddr sexpr))))
       
       ; fget
       ((fget) (make-fget-expr (cadr sexpr)))
       
       ; fset
       ((fset!) (make-fset-expr (cadr sexpr)
                                (parse-expr (caddr sexpr))))
       
       ; send
       ((send) (make-send-expr (parse-expr (cadr sexpr))
                               (caddr sexpr)
                               (parse-exprs (cdddr sexpr))))
       
       ; super
       ((super) (make-super-expr (cadr sexpr)
                                 (parse-exprs (cddr sexpr))))
       
       ; ----------------------------------------------------------------
       ; thread
       ((thread) (make-thread-expr (parse-expr (cadr sexpr))))
       
       ; wait 
       ((wait) (make-wait-expr (cadr sexpr)))
       
       ; notifyAll
       ((notify) (make-notify-expr (cadr sexpr)))
       
       ; ----------------------------------------------------------------
       ; deploy
       ((deploy) (make-deploy-expr (map (lambda(asp)
                                          (make-dyn-aspect (parse-expr (car asp)) ;c
                                                           (parse-expr (cadr asp)) ;d
                                                           (parse-expr (caddr asp)) ;f
                                                           (parse-expr (cadddr asp)))) ; aspect
                                        (cadr sexpr))
                                   (map parse-expr (cddr sexpr)))) ; body
                 
       ; deploy-on
       ((deploy-on) (make-deploy-on-expr (map (lambda(asp)
                                                (make-dyn-aspect (parse-expr (car asp)) ;c
                                                                 (parse-expr (cadr asp)) ;d
                                                                 (parse-expr (caddr asp)) ;f
                                                                 (parse-expr (cadddr asp)))) ; aspect
                                              (cadr sexpr))
                                         (map parse-expr (cddr sexpr)))) ; body
       
       ; prim-app
       (else
        (let ((prim (assq (car sexpr) *primitives*)))
          (if prim
              (make-prim-app-expr (cadr prim) (parse-exprs (cdr sexpr)))
              (error 'parse-expr "~a~n" sexpr))))
       ))))

(define (prim-lit? s) (foldl (lambda(f v)(or v (f s)))
                             #f 
                             (list number? string? boolean?)))

(define (list-lit? s) (and (list? s) (equal? (car s) 'list)))

;;; primitives
(define *primitives*
  `((+       ,(lambda args (apply + args)))
    (-       ,(lambda args (apply - args)))
    (*       ,(lambda args (apply * args)))
    (/       ,(lambda args (apply / args)))
    (=       ,(lambda args (apply = args)))
    (equal?  ,(lambda args (apply equal? args)))
    (<       ,(lambda args (apply < args)))
    (>       ,(lambda args (apply > args)))
    (zero?   ,(lambda args (apply zero? args)))
    
    (not     ,(lambda args (apply not args)))
    
    (write   ,(lambda args (apply write args)))
    (printf  ,(lambda args (apply printf args)))
    (newline ,(lambda args (apply newline args)))
    
    (car     ,(lambda args (apply car  args)))
    (cdr     ,(lambda args (apply cdr  args)))
    
    (cons    ,(lambda args (apply cons args)))
    (list    ,(lambda args args))
    (length  ,(lambda args (apply length args)))
    (null?   ,(lambda args (apply null? args)))
    (list?   ,(lambda args (apply list? args)))
    
    (number? ,(lambda args (apply number? args)))
    
    ;string
    (substr   ,(lambda args (apply substring args)))
    (str-app  ,(lambda args (apply string-append args)))
    
    (error   ,(lambda args (apply error args)))
    (sleep   ,(lambda args (apply sleep args)))
    (random  ,(lambda args (apply random args)))
    (exit    ,(lambda args (apply exit args)))
    ;(throw   ,(lambda args (apply printf args) (kill-thread (current-thread))))
    (throw   ,(lambda args (apply error (cons 'throw args))))
    
    ; some reflection
    (class-name  ,(lambda(obj) (rt-class-name (rt-object-class obj))))
    
    ;jps
    (jp-kind   ,(lambda(jp) (jp-kind jp)))
    (jp-name   ,(lambda(jp) (jp-name jp)))
    (jp-target ,(lambda(jp) (jp-target jp)))
    (jp-args   ,(lambda(jp) (jp-args jp)))
    
    (jp-str    ,(lambda(jp) (printf "jp: ~a ~a ~a ~a~n"
                                    (jp-kind jp)
                                    (jp-name jp) 
                                    (if (rt-class? (jp-target jp))
                                        (rt-class-name (jp-target jp))
                                        (rt-class-name (rt-object-class (jp-target jp))))
                                    (jp-args jp))))
    
    ; standard poincuts (we can't return lambdas here)
    ; (define-struct jp   (kind name target args parent))
    (call    ,(lambda(jp name)
                (and (equal? (jp-kind jp) 'call)
                     (equal? (jp-name jp) name))))
    
    (call2   ,(lambda(jp cname mname)
                (and (equal? (jp-kind jp) 'call)
                     (equal? (jp-name jp) mname)
                     (equal? (rt-class-name (rt-object-class (jp-target jp))) cname))))
    
    (exec    ,(lambda(jp name)
                (and (equal? (jp-kind jp) 'exec)
                     (equal? (jp-name jp) name))))
    
    (exec2   ,(lambda(jp cname mname)
                (and (equal? (jp-kind jp) 'exec)
                     (equal? (rt-class-name (rt-object-class (jp-target jp))) cname)
                     (equal? (jp-name jp) mname) )))
    
    (creation ,(lambda(jp name)
                 (and (equal? (jp-kind jp) 'new)
                      (equal? (rt-class-name (jp-target jp)) name))))
    
    ; higher-order pointcuts & aux pcs
    (xcall   ,(lambda(name)
                (lambda(jp)
                  (and (equal? (jp-kind jp) 'call) (equal? (jp-name jp) name)))))
    
    (cflow      ,(lambda(jp pc)(cflow jp pc)))
    (cflowbelow ,(lambda(jp pc)(cflowbelow jp pc)))
    ))

(define (cflow jp pc)
  (or (pc jp) (cflowbelow jp pc)))

(define (cflowbelow jp pc)
  (and (jp-parent jp)
       (cflow (jp-parent jp) pc)))